/* Developed for Lenovo Educational Board Pass 2
   © Lenovo Created by Grason Humphrey & Modified by Joshua Grosser
   v. 1.0 */

#include <Wire.h>
#include "SSD1306Wire.h"

SSD1306Wire display(0x3c);
#define WIRE Wire

// Define game constants

// Size of each snake segment
const int BOX_SIZE = 8;
const int MAX_SPEED = 8;
// Width and height of screen
const int SCREEN_WIDTH = 128;
const int SCREEN_HEIGHT = 64;
// Game loop timer, lower value means faster movement
const int TIMER_MAX = 300;
// Location in array for X and Y coordinates
const int X_ARRAY = 0;
const int Y_ARRAY = 1;

int timerCount = TIMER_MAX;

// Direction buttons for snake control
int left = 1;
int down = 2;
int up = 3;
int right = 4;

// Speed of the head of the snake
int xSpeed = 1;
int ySpeed = 0;

// Store location of all snake segments
int snakeMap[128][2];
// Index of random fruit location
int fruitLoc;
// Number of fruits eaten
int snakeSize = 0;
// List of all possible fruit locations, used to keep from spawning fruit on top of snake
int fruitLocs[128][2];
// Current fruit X and Y location
int fruitX;
int fruitY;
// Number of fruit remaining to be eaten before game ends
int fruitRemaining = 127;

void setup() {
  // Start serial communication for debug
  Serial.begin(115200);
  // Start I2C communication with OLED display
  Wire.begin(8, 9);

  // Set snake starting location
  snakeMap[0][X_ARRAY] = 3;
  snakeMap[0][Y_ARRAY] = 3;

  // Find what locations are available for fruit to spawn
  GetGoodFruit();

  fruitLoc = random(0, fruitRemaining);
  fruitX = (int) (fruitLocs[fruitLoc][X_ARRAY]);
  fruitY = (int) (fruitLocs[fruitLoc][Y_ARRAY]);

  // Assign input buttons
  pinMode(left, INPUT);
  pinMode(down, INPUT);
  pinMode(up, INPUT);
  pinMode(right, INPUT);

  pinMode(45, INPUT);

  // I2C display
  display.init();
  display.setContrast(255);

}

// Function to find what spaces are not covered by the snake to spawn fruit
void GetGoodFruit() {
  int numHits = 0;
  for (int i = 0; i < 128; i++) {
    bool snakeHit = false;
    int fruitX = (int) (i / 8);
    int fruitY = (int) (i - (fruitX * 8));
    for (int j = 0; j <= snakeSize; j++) {
      if (snakeMap[j][X_ARRAY] != fruitX || snakeMap[j][Y_ARRAY] != fruitY && !snakeHit) {
      } else if (!snakeHit) {
        ++numHits;
        snakeHit = true;
      }
    }
    if(!snakeHit){
      fruitLocs[i - numHits][X_ARRAY] = fruitX;
      fruitLocs[i - numHits][Y_ARRAY] = fruitY;
    }
  }
}

void loop() {
  // Flag keeps track of if game has ended
  bool gameOver = false;

  // Timer to control game updating speed
  ++timerCount;
  if (timerCount > TIMER_MAX) {
  // Time to update game
  
  // Move snake body forward
    for (int i = snakeSize; i > 0; i--) {
      snakeMap[i][X_ARRAY] = snakeMap[i - 1][X_ARRAY];
      snakeMap[i][Y_ARRAY] = snakeMap[i - 1][Y_ARRAY];
    }
    snakeMap[0][X_ARRAY] += xSpeed;
    snakeMap[0][Y_ARRAY] += ySpeed;

  // Determine if snake has collided with wall
    if (snakeMap[0][X_ARRAY] > 15) {
      snakeMap[0][X_ARRAY] = 15;
      gameOver = true;
    } else if (snakeMap[0][X_ARRAY] < 0) {
      snakeMap[0][X_ARRAY] = 0;
      gameOver = true;
    }
    if (snakeMap[0][Y_ARRAY] > 7) {
      snakeMap[0][Y_ARRAY] = 7;
      gameOver = true;
    } else if (snakeMap[0][Y_ARRAY] < 0) {
      snakeMap[0][Y_ARRAY] = 0;
      gameOver = true;
    }
    
  // Determine if snake has collided with itself
    for(int i = 3; i <= snakeSize; i++){
      if(snakeMap[i][X_ARRAY] == snakeMap[0][X_ARRAY] && snakeMap[i][Y_ARRAY] == snakeMap[0][Y_ARRAY]){
        gameOver = true;
      }
    }

  // Show game over text and score
    if(gameOver){
      display.clear();
      display.setTextAlignment(TEXT_ALIGN_CENTER);
      display.drawString(SCREEN_WIDTH/2, 16, "Game Over");
      display.drawString(SCREEN_WIDTH/2, 32, String("Score: ") + String(snakeSize));
      display.display();
      while (true) {
        delay(10);  
      }
    }

  // Determine if fruit has been eaten
    if (snakeMap[0][X_ARRAY] == fruitX && snakeMap[0][Y_ARRAY] == fruitY) {
      ++snakeSize;
      --fruitRemaining;

      // Show win text if all possible fruit has been eaten
      if(fruitRemaining <= 0){
      display.clear();
      display.setTextAlignment(TEXT_ALIGN_CENTER);
      display.drawString(SCREEN_WIDTH/2, 16, "WOW, You Win!!");
      display.drawString(SCREEN_WIDTH/2, 48, String("Score: ") + String(snakeSize));
      display.display();
      while (true) {
        delay(10);  
      }
    }

      // Find possible fruit locations
      GetGoodFruit();

      fruitLoc = random(0, fruitRemaining);
      fruitX = (int) (fruitLocs[fruitLoc][X_ARRAY]);
      fruitY = (int) (fruitLocs[fruitLoc][Y_ARRAY]);

      // Use snake head forward
      snakeMap[snakeSize][X_ARRAY] = snakeMap[snakeSize - 1][X_ARRAY];
      snakeMap[snakeSize][Y_ARRAY] = snakeMap[snakeSize - 1][Y_ARRAY];
    }

    // Display
    display.clear();
    // Draw snake
    for (int i = 0; i <= snakeSize; i++) {
      display.fillRect(snakeMap[i][X_ARRAY] * BOX_SIZE, snakeMap[i][Y_ARRAY] * BOX_SIZE, BOX_SIZE, BOX_SIZE);
    }
    // Draw fruit
    display.fillCircle(fruitX * BOX_SIZE + BOX_SIZE / 2, fruitY * BOX_SIZE + BOX_SIZE / 2, BOX_SIZE / 2);
    // Draw screen outline
    display.drawRect(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
    display.display();
    timerCount = 0;
  }

  // Get button input
  if (digitalRead(left) && xSpeed == 0) {
    xSpeed = -1;
    ySpeed = 0;
    timerCount = TIMER_MAX;
  } else if (digitalRead(right) && xSpeed == 0) {
    xSpeed = 1;
    ySpeed = 0;
    timerCount = TIMER_MAX;
  }
  else if (digitalRead(up) && ySpeed == 0) {
    ySpeed = -1;
    xSpeed = 0;
    timerCount = TIMER_MAX;
  } else if (digitalRead(down) && ySpeed == 0) {
    ySpeed = 1;
    xSpeed = 0;
    timerCount = TIMER_MAX;
  }
  delay(1);
}
